@@ -17,6 +17,9 @@
 #[cfg(feature = "alloc")]
 extern crate alloc;
 
+pub mod either;
+#[cfg(feature = "alloc")]
+pub mod erased;
 mod is_covariant;
 mod macro_impls;
 pub mod trait_hack;
