@@ -2,6 +2,8 @@
 // called LICENSE at the top level of the ICU4X source tree
 // (online at: https://github.com/unicode-org/icu4x/blob/main/LICENSE ).
 
+#[cfg(feature = "alloc")]
+use crate::erased::{ErasedBoxCart, ErasedRcCart};
 use crate::trait_hack::YokeTraitHack;
 use crate::IsCovariant;
 use crate::Yokeable;
@@ -9,6 +11,8 @@ use core::marker::PhantomData;
 use core::ops::Deref;
 use stable_deref_trait::StableDeref;
 
+#[cfg(feature = "alloc")]
+use alloc::boxed::Box;
 #[cfg(feature = "alloc")]
 use alloc::rc::Rc;
 #[cfg(feature = "alloc")]
@@ -24,7 +28,9 @@ use alloc::sync::Arc;
 /// not the actual lifetime of the data, rather it is a convenient way to erase
 /// the lifetime and make it dynamic.
 ///
-/// `C` is the "cart", which `Y` may contain references to.
+/// `C` is the "cart", which `Y` may contain references to. After the yoke is constructed,
+/// the cart serves little purpose except to guarantee that `Y`'s references remain valid
+/// for as long as the yoke remains in memory (by calling the destructor at the appropriate moment).
 ///
 /// The primary constructor for [`Yoke`] is [`Yoke::attach_to_cart()`]. Several variants of that
 /// constructor are provided to serve numerous types of call sites and `Yoke` signatures.
@@ -205,6 +211,7 @@ impl<Y: for<'a> Yokeable<'a>, C> Yoke<Y, C> {
     /// let yoke: Yoke<Cow<str>, _> = load_object("filename.bincode");
     /// assert_eq!(yoke.get(), "hello");
     /// ```
+    #[inline]
     pub fn get<'a>(&'a self) -> &'a <Y as Yokeable<'a>>::Output {
         self.yokeable.transform()
     }
@@ -234,7 +241,7 @@ impl<Y: for<'a> Yokeable<'a>, C> Yoke<Y, C> {
     /// let yoke = Yoke::<
     ///     &'static str,
     ///     Box<String>
-    /// >::attach_to_box_cart(Box::new(local_data));
+    /// >::attach_to_zero_copy_cart(Box::new(local_data));
     /// assert_eq!(*yoke.get(), "foo");
     ///
     /// // Get back the cart
@@ -252,7 +259,7 @@ impl<Y: for<'a> Yokeable<'a>, C> Yoke<Y, C> {
     /// let mut yoke = Yoke::<
     ///     Cow<'static, str>,
     ///     Box<String>
-    /// >::attach_to_box_cart(Box::new(local_data));
+    /// >::attach_to_zero_copy_cart(Box::new(local_data));
     /// assert_eq!(yoke.get(), "foo");
     ///
     /// // Override data in the cart
@@ -271,6 +278,28 @@ impl<Y: for<'a> Yokeable<'a>, C> Yoke<Y, C> {
         self.cart
     }
 
+    /// Unsafe function for replacing the cart with another
+    ///
+    /// This can be used for type-erasing the cart, for example.
+    ///
+    /// # Safety
+    ///
+    /// - `f()` must not panic
+    /// - References from the yokeable `Y` should still be valid for the lifetime of the
+    ///   returned cart type.
+    ///
+    /// Typically, this means implementing `f` as something which _wraps_ the inner cart type.
+    /// `Yoke` only really cares about destructors for its carts so it's fine to erase other
+    /// information about the cart, as long as the backing data will still be destroyed at the
+    /// same time.
+    #[inline]
+    pub unsafe fn replace_cart<C2>(self, f: impl FnOnce(C) -> C2) -> Yoke<Y, C2> {
+        Yoke {
+            yokeable: self.yokeable,
+            cart: f(self.cart),
+        }
+    }
+
     /// Mutate the stored [`Yokeable`] data.
     ///
     /// See [`Yokeable::transform_mut()`] for why this operation is safe.
@@ -360,6 +389,15 @@ impl<Y: for<'a> Yokeable<'a>, C> Yoke<Y, C> {
     {
         self.yokeable.transform_mut(f)
     }
+
+    /// Helper function allowing one to wrap the cart type in an `Option<T>`.
+    #[inline]
+    pub fn wrap_cart_in_option(self) -> Yoke<Y, Option<C>> {
+        unsafe {
+            // safe because the cart is preserved, just wrapped
+            self.replace_cart(Some)
+        }
+    }
 }
 
 impl<Y: for<'a> Yokeable<'a>> Yoke<Y, ()> {
@@ -388,6 +426,15 @@ impl<Y: for<'a> Yokeable<'a>> Yoke<Y, ()> {
     pub fn new_always_owned(yokeable: Y) -> Self {
         Self { yokeable, cart: () }
     }
+
+    /// Obtain the yokeable out of a `Yoke<Y, ()>`
+    ///
+    /// For most `Yoke` types this would be unsafe but it's
+    /// fine for `Yoke<Y, ()>` since there are no actual internal
+    /// references
+    pub fn into_yokeable(self) -> Y {
+        self.yokeable
+    }
 }
 
 impl<Y: for<'a> Yokeable<'a>, C: StableDeref> Yoke<Y, Option<C>> {
@@ -396,7 +443,7 @@ impl<Y: for<'a> Yokeable<'a>, C: StableDeref> Yoke<Y, Option<C>> {
     /// this is good for e.g. constructing fully owned
     /// [`Yoke`]s with no internal borrowing.
     ///
-    /// This can be paired with [`Yoke::attach_to_option_cart()`] to mix owned
+    /// This can be paired with [`Yoke:: wrap_cart_in_option()`] to mix owned
     /// and borrowed data.
     ///
     /// If you do not wish to pair this with borrowed data, [`Yoke::new_always_owned()`] can
@@ -422,36 +469,14 @@ impl<Y: for<'a> Yokeable<'a>, C: StableDeref> Yoke<Y, Option<C>> {
         }
     }
 
-    /// Similar to [`Yoke::attach_to_cart()`], except it constructs a `Yoke<Y, Option<C>>`
-    /// instead, where the cart is `Some(..)`.
+    /// Obtain the yokeable out of a `Yoke<Y, Option<C>>` if possible.
     ///
-    /// This allows mixing [`Yoke`]s constructed from owned and borrowed data, when
-    /// paired with [`Yoke::new_owned()`].
-    ///
-    /// This method is currently unusable due to a [compiler bug](https://github.com/rust-lang/rust/issues/84937),
-    /// use [`Yoke::attach_to_option_cart_badly()`] instead
-    pub fn attach_to_option_cart<F>(cart: C, f: F) -> Self
-    where
-        F: for<'de> FnOnce(&'de <C as Deref>::Target) -> <Y as Yokeable<'de>>::Output,
-    {
-        let deserialized = f(cart.deref());
-        Self {
-            yokeable: unsafe { Y::make(deserialized) },
-            cart: Some(cart),
-        }
-    }
-    /// Temporary version of [`Yoke::attach_to_option_cart()`]
-    /// that doesn't hit <https://github.com/rust-lang/rust/issues/84937>
-    ///
-    /// See its docs for more details
-    pub fn attach_to_option_cart_badly(
-        cart: C,
-        f: for<'de> fn(&'de <C as Deref>::Target) -> <Y as Yokeable<'de>>::Output,
-    ) -> Self {
-        let deserialized = f(cart.deref());
-        Self {
-            yokeable: unsafe { Y::make(deserialized) },
-            cart: Some(cart),
+    /// If the cart is `None`, this returns `Some`, but if the cart is `Some`,
+    /// this returns `self` as an error.
+    pub fn try_into_yokeable(self) -> Result<Y, Self> {
+        match self.cart {
+            Some(_) => Err(self),
+            None => Ok(self.yokeable),
         }
     }
 }
@@ -509,9 +534,9 @@ unsafe impl<'b, Y: for<'a> Yokeable<'a>, C: IsCovariant<'b>> IsCovariant<'b> for
 #[test]
 fn test_clone() {
     let local_data = "foo".to_string();
-    let y1 = Yoke::<alloc::borrow::Cow<'static, str>, Rc<String>>::attach_to_rc_cart(Rc::new(
-        local_data,
-    ));
+    let y1 = Yoke::<alloc::borrow::Cow<'static, str>, Rc<String>>::attach_to_zero_copy_cart(
+        Rc::new(local_data),
+    );
 
     // Test basic clone
     let y2 = y1.clone();
@@ -752,6 +777,123 @@ impl<Y: for<'a> Yokeable<'a>, C> Yoke<Y, C> {
     }
 }
 
+#[cfg(feature = "alloc")]
+impl<Y: for<'a> Yokeable<'a>, C: 'static + Sized> Yoke<Y, Rc<C>> {
+    /// Allows type-erasing the cart in a `Yoke<Y, Rc<C>>`.
+    ///
+    /// The yoke only carries around a cart type for its destructor,
+    /// since it needs to be able to guarantee that its internal references
+    /// are valid for the lifetime of the Yoke. As such, the actual type of the
+    /// Cart is not very useful unless you wish to extract data out of it
+    /// via [`Yoke::backing_cart()`]. Erasing the cart allows for one to mix
+    /// [`Yoke`]s obtained from different sources.
+    ///
+    /// In case the cart type is not already `Rc<T>`, you can use
+    /// [`Yoke::wrap_cart_in_rc()`] to wrap it.
+    ///
+    /// # Example
+    ///
+    /// ```rust
+    /// use yoke::Yoke;
+    /// use yoke::erased::ErasedRcCart;
+    /// use std::rc::Rc;
+    ///
+    /// let buffer1: Rc<String> = Rc::new("   foo bar baz  ".into());
+    /// let buffer2: Box<String> = Box::new("  baz quux  ".into());
+    ///
+    /// let yoke1 = Yoke::<&'static str, _>::attach_to_cart_badly(buffer1, |rc| rc.trim());
+    /// let yoke2 = Yoke::<&'static str, _>::attach_to_cart_badly(buffer2, |b| b.trim());
+    ///
+    ///
+    /// let erased1: Yoke<_, ErasedRcCart> = yoke1.erase_rc_cart();
+    /// // Wrap the Box in an Rc to make it compatible
+    /// let erased2: Yoke<_, ErasedRcCart> = yoke2.wrap_cart_in_rc().erase_rc_cart();
+    ///
+    /// // Now erased1 and erased2 have the same type!
+    /// ```
+    ///
+    /// Available with the `"alloc"` feature enabled.
+    pub fn erase_rc_cart(self) -> Yoke<Y, ErasedRcCart> {
+        unsafe {
+            // safe because the cart is preserved, just
+            // type-erased
+            self.replace_cart(|c| c as ErasedRcCart)
+        }
+    }
+}
+
+#[cfg(feature = "alloc")]
+impl<Y: for<'a> Yokeable<'a>, C: 'static + Sized> Yoke<Y, Box<C>> {
+    /// Allows type-erasing the cart in a `Yoke<Y, Box<C>>`.
+    ///
+    /// The yoke only carries around a cart type for its destructor,
+    /// since it needs to be able to guarantee that its internal references
+    /// are valid for the lifetime of the Yoke. As such, the actual type of the
+    /// Cart is not very useful unless you wish to extract data out of it
+    /// via [`Yoke::backing_cart()`]. Erasing the cart allows for one to mix
+    /// [`Yoke`]s obtained from different sources.
+    ///
+    /// In case the cart type is not already `Box<T>`, you can use
+    /// [`Yoke::wrap_cart_in_box()`] to wrap it.
+    ///
+    /// # Example
+    ///
+    /// ```rust
+    /// use yoke::Yoke;
+    /// use yoke::erased::ErasedBoxCart;
+    /// use std::rc::Rc;
+    ///
+    /// let buffer1: Rc<String> = Rc::new("   foo bar baz  ".into());
+    /// let buffer2: Box<String> = Box::new("  baz quux  ".into());
+    ///
+    /// let yoke1 = Yoke::<&'static str, _>::attach_to_cart_badly(buffer1, |rc| rc.trim());
+    /// let yoke2 = Yoke::<&'static str, _>::attach_to_cart_badly(buffer2, |b| b.trim());
+    ///
+    ///
+    /// // Wrap the Rc in an Box to make it compatible
+    /// let erased1: Yoke<_, ErasedBoxCart> = yoke1.wrap_cart_in_box().erase_box_cart();
+    /// let erased2: Yoke<_, ErasedBoxCart> = yoke2.erase_box_cart();
+    ///
+    /// // Now erased1 and erased2 have the same type!
+    /// ```
+    ///
+    /// Available with the `"alloc"` feature enabled.
+    pub fn erase_box_cart(self) -> Yoke<Y, ErasedBoxCart> {
+        unsafe {
+            // safe because the cart is preserved, just
+            // type-erased
+            self.replace_cart(|c| c as ErasedBoxCart)
+        }
+    }
+}
+
+#[cfg(feature = "alloc")]
+impl<Y: for<'a> Yokeable<'a>, C> Yoke<Y, C> {
+    /// Helper function allowing one to wrap the cart type in a `Box<T>`.
+    /// Can be paired with [`Yoke::erase_box_cart()`]
+    ///
+    /// Available with the `"alloc"` feature enabled.
+    #[inline]
+    pub fn wrap_cart_in_box(self) -> Yoke<Y, Box<C>> {
+        unsafe {
+            // safe because the cart is preserved, just wrapped
+            self.replace_cart(Box::new)
+        }
+    }
+    /// Helper function allowing one to wrap the cart type in an `Rc<T>`.
+    /// Can be paired with [`Yoke::erase_rc_cart()`], or generally used
+    /// to make the [`Yoke`] cloneable.
+    ///
+    /// Available with the `"alloc"` feature enabled.
+    #[inline]
+    pub fn wrap_cart_in_rc(self) -> Yoke<Y, Rc<C>> {
+        unsafe {
+            // safe because the cart is preserved, just wrapped
+            self.replace_cart(Rc::new)
+        }
+    }
+}
+
 /// Safety docs for project()
 ///
 /// (Docs are on a private const to allow the use of compile_fail doctests)
